@@ -14,6 +14,7 @@
 # limitations under the License.
 """Importer sources."""
 
+import json
 import hashlib
 import logging
 import os
@@ -31,26 +32,8 @@ AUTHOR_EMAIL = 'infra@osv.dev'
 PUSH_RETRIES = 2
 PUSH_RETRY_SLEEP_SECONDS = 10
 
-
-class GitRemoteCallback(pygit2.RemoteCallbacks):
-  """Authentication callbacks."""
-
-  def __init__(self, username, ssh_key_public_path, ssh_key_private_path):
-    super().__init__()
-    self._username = username
-    self._ssh_key_public_path = ssh_key_public_path
-    self._ssh_key_private_path = ssh_key_private_path
-
-  def credentials(self, url, username_from_url, allowed_types):
-    logging.info('Allowed types = %s\n', allowed_types)
-    if allowed_types & pygit2.credentials.GIT_CREDENTIAL_USERNAME:
-      return pygit2.Username(self._username)
-
-    if allowed_types & pygit2.credentials.GIT_CREDENTIAL_SSH_KEY:
-      return pygit2.Keypair(self._username, self._ssh_key_public_path,
-                            self._ssh_key_private_path, '')
-
-    return None
+YAML_EXTENSIONS = ('.yaml', '.yml')
+JSON_EXTENSIONS = ('.json')
 
 
 def parse_source_id(source_id):
@@ -64,20 +47,66 @@ def repo_path(repo):
   return os.path.dirname(repo.path.rstrip(os.sep))
 
 
-def parse_vulnerability(path):
-  """Parse vulnerability YAML."""
-  vulnerability = vulnerability_pb2.Vulnerability()
+def _parse_vulnerability_dict(path):
+  """Parse a vulnerability file into a dict."""
   with open(path) as f:
-    data = yaml.safe_load(f)
-  json_format.ParseDict(data, vulnerability)
+    ext = os.path.splitext(path)[1]
+    if ext in YAML_EXTENSIONS:
+      return yaml.safe_load(f)
 
-  return vulnerability
+    if ext in JSON_EXTENSIONS:
+      return json.load(f)
+
+    raise RuntimeError('Unknown format ' + ext)
+
+  return None
 
 
-def parse_vulnerability_from_dict(data):
+def parse_vulnerability(path, key_path=None):
+  """Parse vulnerability YAML."""
+  data = _parse_vulnerability_dict(path)
+  return parse_vulnerability_from_dict(data, key_path)
+
+
+def _parse_vulnerabilities(data, key_path):
+  """Parse multiple vulnerabilities."""
+  if isinstance(data, list):
+    return [parse_vulnerability_from_dict(v, key_path) for v in data]
+
+  return [parse_vulnerability_from_dict(data, key_path)]
+
+
+def parse_vulnerabilities(path, key_path=None):
+  """Parse vulnerabilities (potentially multiple in a list)."""
+  return _parse_vulnerabilities(_parse_vulnerability_dict(path), key_path)
+
+
+def parse_vulnerabilities_from_data(data, extension, key_path=None):
+  """Parse vulnerabilities from data."""
+  if extension in YAML_EXTENSIONS:
+    data = yaml.safe_load(data)
+  elif extension in JSON_EXTENSIONS:
+    data = json.loads(data)
+  else:
+    raise RuntimeError('Unknown format ' + extension)
+
+  return _parse_vulnerabilities(data, key_path)
+
+
+def _get_nested_vulnerability(data, key_path=None):
+  """Get nested vulnerability."""
+  if key_path:
+    for component in key_path.split('.'):
+      data = data[component]
+
+  return data
+
+
+def parse_vulnerability_from_dict(data, key_path=None):
   """Parse vulnerability from dict."""
+  data = _get_nested_vulnerability(data, key_path)
   vulnerability = vulnerability_pb2.Vulnerability()
-  json_format.ParseDict(data, vulnerability)
+  json_format.ParseDict(data, vulnerability, ignore_unknown_fields=True)
   return vulnerability
 
 
@@ -101,11 +130,35 @@ def vulnerability_to_dict(vulnerability):
       vulnerability, preserving_proto_field_name=True)
 
 
-def vulnerability_to_yaml(vulnerability, output_path):
-  """Convert Vulnerability to YAML."""
-  with open(output_path, 'w') as handle:
-    data = vulnerability_to_dict(vulnerability)
-    yaml.dump(data, handle, sort_keys=False, Dumper=YamlDumper)
+def _write_vulnerability_dict(data, output_path):
+  """Write a vulnerability dict to disk."""
+  with open(output_path, 'w') as f:
+    ext = os.path.splitext(output_path)[1]
+    if ext in YAML_EXTENSIONS:
+      yaml.dump(data, f, sort_keys=False, Dumper=YamlDumper)
+    elif ext in JSON_EXTENSIONS:
+      json.dump(data, f, indent=2)
+    else:
+      raise RuntimeError('Unknown format ' + ext)
+
+
+def write_vulnerability(vulnerability, output_path, key_path=None):
+  """Update a vulnerability file on disk."""
+  if os.path.exists(output_path):
+    data = _parse_vulnerability_dict(output_path)
+  else:
+    # Set up the expected nesting based on key_path.
+    data = {}
+    if key_path:
+      cur = data
+      for component in key_path.split('.'):
+        cur[component] = {}
+        cur = cur[component]
+
+  vuln_data = _get_nested_vulnerability(data, key_path)
+  vuln_data.clear()
+  vuln_data.update(vulnerability_to_dict(vulnerability))
+  _write_vulnerability_dict(data, output_path)
 
 
 def vulnerability_has_range(vulnerability, introduced, fixed):
